from typing import NamedTuple
from math import radians, cos, sin
import numpy as np
import xarray as xr
import sasktran as sk
from osirisl1services.spatialrotation import Vector, UnitQuaternion, TaitBryanZYX
from osirisl1services.readlevel1 import open_level1_spectrograph, limit_to_scan, Orbit


class PixelParams(NamedTuple):
    dy: float
    dz: float
    y0: float
    z0: float


# Spectrograph alignment data from OsirisConfigurationDatabase/flightduration/osoptics.ini
# The rotation is expressed as a euler angles "zyx" in degrees.
# 2005-01-03 NDL Initialize with the Jupiter alignment measurements.
SPECTROGRAPH_ROTATION_TB_ZYX = TaitBryanZYX(roll=radians(0.0337), pitch=radians(-0.0903), yaw=radians(-0.15810304))
SPECTROGRAPH_ROTATION_CF = SPECTROGRAPH_ROTATION_TB_ZYX.to_quaternion()


# InfraRed alignment data from OsirisConfigurationDatabase/flightduration/iroptics.ini
# The rotation is expressed as a euler angles "zyx" in degrees.
IR_ROTATION_TB_ZYX = TaitBryanZYX(roll=radians(0.0), pitch=radians(-0.06167), yaw=radians(0.0))
IR_ROTATION_CF = IR_ROTATION_TB_ZYX.to_quaternion()
MINUTES_TO_RADIANS = radians(1.0 / 60.0)
# IR1 Optics
IR1_PIXEL_PARAMS = PixelParams(dy=-1.1931386 * MINUTES_TO_RADIANS, dz=0.0, y0=105.104, z0=0.0)
# IR2 Optics
IR2_PIXEL_PARAMS = PixelParams(dy=-1.2046809 * MINUTES_TO_RADIANS, dz=0.0, y0=109.289, z0=0.0)
# IR3 Optics
IR3_PIXEL_PARAMS = PixelParams(dy=-1.2059574 * MINUTES_TO_RADIANS, dz=0.0, y0=109.492, z0=0.0)


def get_spectrograph_pointing_for_attitude_dataset(quaternion: xr.DataArray) -> xr.DataArray:
    """ Calculate spectrograph look vectors for a DataArray of Quaternions.
        The final rotation is the combined rotation of spacecraft and spectrograph offset:

           q_total = q_spacecraft * q_spectrograph_rotation_cf
    """
    pointing = [(UnitQuaternion(*q) * SPECTROGRAPH_ROTATION_CF).rot_unit_x() for q in quaternion.values]
    return xr.DataArray(pointing, dims=['mjd', 'xyz'])


def get_spectrograph_pointing(mjd: float, quaternion: xr.DataArray) -> Vector:
    """ Calculate spectrograph look vectors for an mjd and attitude Dataset.
        The final rotation is the combined rotation of spacecraft and spectrograph offset:

           q_total = q_spacecraft * q_spectrograph_rotation_cf
    """
    index = np.searchsorted(quaternion.mjd.values, mjd)
    alpha = (mjd - quaternion.mjd[index - 1]) / (quaternion.mjd[index] - quaternion.mjd[index - 1])
    q0 = UnitQuaternion(*quaternion.values[index - 1])
    q1 = UnitQuaternion(*quaternion.values[index])
    return (q0.slerp(q1, alpha) * SPECTROGRAPH_ROTATION_CF).rot_unit_x()


def get_pointing_for_spectrograph_dataset(spectrograph: xr.Dataset,
                                          quaternion: xr.DataArray) -> xr.DataArray:
    """ Calculate spectrograph look vectors for the mjds in a spectrograph Dataset.
        The final rotation is the combined rotation of spacecraft and spectrograph offset:

           q_total = q_spacecraft * q_spectrograph_rotation_cf
    """
    pointing = []
    for mjd in spectrograph.mjd:
        mjd = float(mjd)
        index = np.searchsorted(quaternion.mjd.values, mjd)
        alpha = (mjd - quaternion.mjd[index - 1]) / (quaternion.mjd[index] - quaternion.mjd[index - 1])
        q0 = UnitQuaternion(*quaternion.values[index - 1])
        q1 = UnitQuaternion(*quaternion.values[index])
        pointing.append((q0.slerp(q1, alpha) * SPECTROGRAPH_ROTATION_CF).rot_unit_x())
    return xr.DataArray(pointing, coords=[spectrograph.mjd, ['x', 'y', 'z']], dims=['mjd', 'xyz'])


def get_position_for_spectrograph_dataset(spectrograph: xr.Dataset, attitude: xr.Dataset) -> xr.DataArray:
    """ Calculate satellite position for the mjds in a spectrograph Dataset. """
    mjd_orbit_state = attitude['mjdorbitstate']
    index = np.searchsorted(mjd_orbit_state, spectrograph.mjd)
    future = attitude.isel(mjd=index)
    past = attitude.isel(mjd=index - 1)
    past['mjd'] = future['mjd'] = spectrograph.drop(['stw']).mjd  # TODO: Can this be done some other way?
    alpha = (spectrograph.mjd - past.mjdorbitstate) / (future.mjdorbitstate - past.mjdorbitstate)
    position = past.position * (1.0 - alpha) + future.position * alpha
    return position


def get_tangent_point_data_for_spectrograph_dataset(spectrograph: xr.Dataset,
                                                    attitude: xr.Dataset) -> xr.Dataset:
    """ Calculate spectrograph tangent information (lat, lon, altitude) for the mjds in a spectrograph Dataset. """
    if 'look_eci' not in spectrograph.keys():
        spectrograph['look_eci'] = get_pointing_for_spectrograph_dataset(spectrograph, attitude['Qachieved'])
    if 'position_eci' not in spectrograph.keys():
        spectrograph['position_eci'] = get_position_for_spectrograph_dataset(spectrograph, attitude)

    latitude = []
    longitude = []
    altitude = []
    from osirisl1services.spatialrotation import ECI
    for mjd, spect in spectrograph.groupby('mjd'):
        tangentpoint = sk.LineOfSight(mjd=mjd,
                                      observer=ECI(*spect['position_eci'].values).to_geo(mjd),
                                      look_vector=ECI(*spect['look_eci'].values).to_geo(mjd)).tangent_location()
        if tangentpoint:
            latitude.append(tangentpoint.latitude)
            longitude.append(tangentpoint.longitude)
            altitude.append(tangentpoint.altitude)
        else:
            latitude.append(np.NaN)
            longitude.append(np.NaN)
            altitude.append(np.NaN)

    ds = xr.Dataset({'latitude': (['mjd'], latitude),
                     'longitude': (['mjd'], longitude),
                     'altitude': (['mjd'], altitude)},
                    coords={'mjd': spectrograph.mjd})
    return ds


def get_pointing_for_ir_dataset(channel: int, ir: xr.Dataset, quaternion: xr.DataArray) -> xr.DataArray:
    """ Calculate IR look vectors for the mjds in a ir Dataset.
        The final rotation is the combined rotation of spacecraft and spectrograph offset:

           q_total = q_spacecraft * q_spectrograph_rotation_cf
    """
    pixel_dict = {1: IR1_PIXEL_PARAMS, 2: IR2_PIXEL_PARAMS, 3: IR3_PIXEL_PARAMS}
    pointing = []
    thetas = (ir.pixel + 0.5 - pixel_dict[channel].y0) * pixel_dict[channel].dy
    pixel_cfs = [Vector(cos(theta), sin(theta), 0.0) for theta in thetas]
    for mjd in ir.mjd:
        index = np.searchsorted(quaternion.mjd.values, mjd)
        alpha = (mjd - quaternion.mjd[index - 1]) / (quaternion.mjd[index] - quaternion.mjd[index - 1])
        q0 = UnitQuaternion(*quaternion.values[index - 1])
        q1 = UnitQuaternion(*quaternion.values[index])
        cf_pointing = q0.slerp(q1, alpha)
        ir_pointing = cf_pointing * IR_ROTATION_CF
        pointing.append(ir_pointing.rot_no_z(pixel_cfs))
    pointing = np.reshape(pointing, (len(ir.mjd), 128, 3))
    return xr.DataArray(np.array(pointing), coords=[ir.mjd, range(128), ['x', 'y', 'z']], dims=['mjd', 'pixel', 'xyz'])


def get_position_for_ir_dataset(ir: xr.Dataset, attitude: xr.Dataset) -> xr.DataArray:
    """ Calculate satellite position for the mjds in a ir Dataset. """
    mjd_orbit_state = attitude['mjdorbitstate']
    index = np.searchsorted(mjd_orbit_state, ir.mjd)
    future = attitude.isel(mjd=index)
    past = attitude.isel(mjd=index - 1)
    past['mjd'] = future['mjd'] = ir.mjd
    alpha = (ir.mjd - past.mjdorbitstate) / (future.mjdorbitstate - past.mjdorbitstate)
    position = past.position * (1.0 - alpha) + future.position * alpha
    return position


def get_tangent_point_data_for_ir_dataset(channel: int, ir: xr.Dataset, attitude: xr.Dataset) -> xr.Dataset:
    """ Calculate IR tangent information (lat, lon, altitude) for the mjds in a ir Dataset. """
    if 'look_eci' not in ir.keys():
        ir['look_eci'] = get_pointing_for_ir_dataset(channel, ir, attitude['Qachieved'])
    if 'position_eci' not in ir.keys():
        ir['position_eci'] = get_position_for_ir_dataset(ir, attitude)

    latitude = []
    longitude = []
    altitude = []
    from osirisl1services.spatialrotation import ECI
    for mjd, ir1 in ir.groupby('mjd'):
        for pixel, ir2 in ir1.groupby('pixel'):
            tangentpoint = sk.LineOfSight(mjd=mjd,
                                          observer=ECI(*ir1['position_eci'].values).to_geo(mjd),
                                          look_vector=ECI(*ir2['look_eci'].values).to_geo(mjd)).tangent_location()
            if tangentpoint:
                latitude.append(tangentpoint.latitude)
                longitude.append(tangentpoint.longitude)
                altitude.append(tangentpoint.altitude)
            else:
                latitude.append(np.NaN)
                longitude.append(np.NaN)
                altitude.append(np.NaN)

    latitude = np.reshape(latitude, (len(ir.mjd), 128))
    longitude = np.reshape(longitude, (len(ir.mjd), 128))
    altitude = np.reshape(altitude, (len(ir.mjd), 128))

    ds = xr.Dataset({'latitude': (['mjd', 'pixel'], latitude),
                     'longitude': (['mjd', 'pixel'], longitude),
                     'altitude': (['mjd', 'pixel'], altitude)},
                    coords={'mjd': ir.mjd,
                            'pixel': range(128)})
    return ds


def get_scans_in_orbit_below_sza(orbit: Orbit, sza: float, target_altitude: float = 25000):
    """ For a given orbit return an iterator of the scans whose tangent point is above a configurable horizon.
        Useful for retrieval code. """
    from osirisl1services.readlevel1 import open_level1_attitude
    from astropy.time import Time
    from astropy import units as u
    from astropy.coordinates import EarthLocation, AltAz
    from astropy.coordinates import get_sun

    try:
        attitude = open_level1_attitude(orbit=orbit, load_pointing=True)
        spectrograph = open_level1_spectrograph(orbit=orbit)
    except LookupError:  # No valid orbits
        raise StopIteration

    for scanno in attitude.ScanNumber.values % 1000:
        spect = limit_to_scan(spectrograph, orbit, scanno)
        tp = get_tangent_point_data_for_spectrograph_dataset(spect, attitude)\
            .swap_dims({'mjd': 'altitude'}).sortby('altitude').sel(altitude=target_altitude, method='nearest')

        location = EarthLocation(lat=tp.latitude * u.deg, lon=tp.longitude * u.deg)

        time = Time(tp.mjd, format='mjd')

        sun = get_sun(time).transform_to(AltAz(obstime=time, location=location))

        if sun.alt > (90.0 - sza) * u.deg:
            yield scanno
