from unittest import TestCase
import numpy as np
from osirisl1services.readlevel1 import Orbit, open_level1_attitude, open_level1_spectrograph, open_level1_ir
import osl1 as l1
from osirisl1services.spatialrotation import ECI, ECEF, Geodetic
import sasktran as sk
from osirisl1services.services import (get_spectrograph_pointing_for_attitude_dataset, get_spectrograph_pointing,
                                       get_pointing_for_spectrograph_dataset,
                                       get_position_for_spectrograph_dataset,
                                       get_tangent_point_data_for_spectrograph_dataset,
                                       get_pointing_for_ir_dataset, get_position_for_ir_dataset,
                                       get_tangent_point_data_for_ir_dataset,
                                       get_scans_in_orbit_below_sza,
                                       SPECTROGRAPH_ROTATION_TB_ZYX, IR_ROTATION_TB_ZYX)


class TestLevel1Services(TestCase):
    def setUp(self):
        self.scan_list = [6432013,  # catches #25
                          27110030]  # 27110030 catches #25

    def test_spectrograph_look_cf(self):
        """ Test that the look direction from c++ Level1 Services is the same as derived by the TB angles in
            /OsirisConfigurationDatabase/flightduration/osoptics.ini

            Vector(x=0.99999495087295287, y=-0.0027594116765072515, z=0.0015760316621056657)
        """
        osiris_look_cf = l1.osl1.GetCFUnitVector(55000.0, l1.osl1.OS_X_BORE)[1]
        q = SPECTROGRAPH_ROTATION_TB_ZYX.to_quaternion()
        np.testing.assert_allclose(q.rot_unit_x(), osiris_look_cf)

    def test_level1_spectrograph_pointing(self):
        """ Test comparing the look vector solution based on the attitude quaternion data to the look vectors returned
            by the c++ Level1 Services.
        """
        for scanno in self.scan_list:
            attitude = open_level1_attitude(Orbit(scanno // 1000), scan=scanno, load_pointing=True)
            attitude['osiris_look_eci'] = get_spectrograph_pointing_for_attitude_dataset(attitude['Qachieved'])
            for mjd, att in attitude.groupby('mjd'):
                osiris_look_cf = l1.osl1.GetCFUnitVector(mjd, l1.osl1.OS_X_BORE)[1]
                osiris_look_eci = l1.osl1.CFToECI(mjd, osiris_look_cf, l1.osl1.ECI_SPACECRAFT)[1]
                np.testing.assert_allclose(att['osiris_look_eci'], osiris_look_eci, atol=1e-6,
                                           err_msg=f'ScanNo: {scanno}', verbose=True)

    def test_spectrograph_pointing_with_interpolation(self):
        """ Test comparing the look vector solution based on the attitude quaternion data to the look vectors returned
            by the c++ Level1 Services on points between the attitude solution.
        """
        for scanno in self.scan_list:
            attitude = open_level1_attitude(Orbit(scanno // 1000), scan=scanno, load_pointing=True)
            for mjd in np.linspace(attitude.mjd.min(), attitude.mjd.max(), 10, endpoint=True):
                osiris_look_cf = l1.osl1.GetCFUnitVector(mjd, l1.osl1.OS_X_BORE)[1]
                osiris_look_eci = l1.osl1.CFToECI(mjd, osiris_look_cf, l1.osl1.ECI_SPACECRAFT)[1]
                look = get_spectrograph_pointing(mjd, attitude['Qachieved'])
                np.testing.assert_allclose(look, osiris_look_eci, atol=1e-6,
                                           err_msg=f'ScanNo: {scanno}', verbose=True)

    def test_pointing_for_spectrograph_dataset(self):
        """ Test comparing the look vector solution based on the attitude quaternion data to the look vectors returned
            by the c++ Level1 Services on points between the attitude solution for a dataset.
        """
        for scanno in self.scan_list:
            attitude = open_level1_attitude(Orbit(scanno // 1000), scan=scanno, load_pointing=True)
            spectrograph = open_level1_spectrograph(Orbit(scanno // 1000), scan=scanno)
            spectrograph['look_eci'] = get_pointing_for_spectrograph_dataset(spectrograph,
                                                                             attitude['Qachieved'])
            scan_info_swig, _, _, _, _ = l1.LoadScanOS(scanno)
            for info in scan_info_swig:
                mjd = info.mjd
                osiris_look_cf = l1.osl1.GetCFUnitVector(mjd, l1.osl1.OS_X_BORE)[1]
                osiris_look_eci = l1.osl1.CFToECI(mjd, osiris_look_cf, l1.osl1.ECI_SPACECRAFT)[1]
                np.testing.assert_allclose(spectrograph['look_eci'].sel(mjd=mjd), osiris_look_eci, atol=1e-6,
                                           err_msg=f'ScanNo: {scanno}', verbose=True)

    def test_level1_spectrograph_SSC_tangent_point(self):
        """ Test comparing the tangent point derived by the attitude file to the tangent point returned
            by the c++ Level1 Services.
        """
        for scanno in self.scan_list:
            attitude = open_level1_attitude(Orbit(scanno // 1000), scan=scanno, load_pointing=True)
            attitude['osiris_look_eci'] = get_spectrograph_pointing_for_attitude_dataset(attitude['Qachieved'])
            for mjd, att in attitude.groupby('mjd'):
                position_eci = ECI(*l1.osl1.GetOdinPosition(mjd)[1])
                position_ecef = ECEF(*l1.osl1.ECIToGEO(mjd, list(position_eci))[1])
                osiris_look_cf = l1.osl1.GetCFUnitVector(mjd, l1.osl1.OS_X_BORE)[1]
                osiris_look_eci = ECI(*l1.osl1.CFToECI(mjd, osiris_look_cf, l1.osl1.ECI_SPACECRAFT)[1])
                osiris_look_ecef = ECEF(*l1.osl1.ECIToGEO(mjd, list(osiris_look_eci))[1])
                tangentpoint_eci = ECI(*l1.osl1.LOSTangentPoint(list(position_eci), list(osiris_look_eci), False)[1])
                tangentpoint = Geodetic(*l1.osl1.ECIToGeodetic(mjd, list(tangentpoint_eci))[1])

                tangent_point_sasktran = sk.LineOfSight(mjd=mjd,
                                                        observer=position_ecef,
                                                        look_vector=osiris_look_ecef).tangent_location()

                self.assertAlmostEqual(tangentpoint.latitude, tangent_point_sasktran.latitude, delta=1e-6,
                                       msg=f'ScanNo: {scanno}')
                self.assertAlmostEqual(tangentpoint.longitude, tangent_point_sasktran.longitude, delta=1e-6,
                                       msg=f'ScanNo: {scanno}')
                self.assertAlmostEqual(tangentpoint.height * 1000.0, tangent_point_sasktran.altitude, delta=1e-6,
                                       msg=f'ScanNo: {scanno}')

    def test_get_tangent_point_data_for_spectrograph_dataset(self):
        """ Test comparing the tangent point solution based on the attitude and look data with the c++ Level1 Services
            on points between the attitude solution for the spectrograph.
        """
        for scanno in self.scan_list:
            attitude = open_level1_attitude(Orbit(scanno // 1000), scan=scanno, load_pointing=True)
            spectrograph = open_level1_spectrograph(Orbit(scanno // 1000), scan=scanno)
            # Load the look vector and transform it to ECI
            spectrograph['look_eci'] = get_pointing_for_spectrograph_dataset(spectrograph, attitude['Qachieved'])
            spectrograph['position_eci'] = get_position_for_spectrograph_dataset(spectrograph, attitude)
            spectrograph.merge(get_tangent_point_data_for_spectrograph_dataset(spectrograph, attitude), inplace=True)
            for mjd, spect in spectrograph.groupby('mjd'):
                position_eci = ECI(*l1.osl1.GetOdinPosition(mjd)[1])
                # Assert the ECI positions are the to 5 meters
                np.testing.assert_allclose(spect['position_eci'], position_eci, atol=5)
                osiris_look_cf = l1.osl1.GetCFUnitVector(mjd, l1.osl1.OS_X_BORE)[1]
                osiris_look_eci = ECI(*l1.osl1.CFToECI(mjd, osiris_look_cf, l1.osl1.ECI_SPACECRAFT)[1])
                tangentpoint_eci = ECI(*l1.osl1.LOSTangentPoint(list(position_eci),
                                                                list(osiris_look_eci), False)[1])
                tangentpoint = Geodetic(*l1.osl1.ECIToGeodetic(mjd, list(tangentpoint_eci))[1])

                self.assertAlmostEqual(float(spect['latitude']), tangentpoint.latitude, delta=0.01,
                                       msg=f'ScanNo: {scanno}')
                self.assertAlmostEqual(float(spect['longitude']), tangentpoint.longitude, delta=0.01,
                                       msg=f'ScanNo: {scanno}')
                self.assertAlmostEqual(float(spect['altitude']), tangentpoint.height * 1000.0, delta=5,
                                       msg=f'ScanNo: {scanno}')

    def test_ir_look_cf(self):
        """ Test that the look direction from c++ Level1 Services is the same as derived by the TB angles in
            /OsirisConfigurationDatabase/flightduration/iroptics.ini

            Vector(x=0.999999420741261, y=0.0, z=0.0010763443418775663)
        """
        ir1_look_cf = l1.osl1.GetCFUnitVector(55000.0, l1.osl1.IR1_X_BORE)[1]
        ir2_look_cf = l1.osl1.GetCFUnitVector(55000.0, l1.osl1.IR2_X_BORE)[1]
        ir3_look_cf = l1.osl1.GetCFUnitVector(55000.0, l1.osl1.IR3_X_BORE)[1]
        q = IR_ROTATION_TB_ZYX.to_quaternion()
        np.testing.assert_allclose(q.rot_unit_x(), ir1_look_cf)
        np.testing.assert_allclose(q.rot_unit_x(), ir2_look_cf)
        np.testing.assert_allclose(q.rot_unit_x(), ir3_look_cf)

    def pointing_for_ir_channel_dataset(self, channel: int):
        """ Test comparing the look vector solution based on the attitude quaternion data to the look vectors returned
            by the c++ Level1 Services on points between the attitude solution for a dataset.
        """
        l1_inst_dict = {1: l1.osl1.INST_IR1, 2: l1.osl1.INST_IR2, 3: l1.osl1.INST_IR3}
        for scanno in self.scan_list:
            attitude = open_level1_attitude(Orbit(scanno // 1000), scan=scanno, load_pointing=True)
            ir = open_level1_ir(Orbit(scanno // 1000), scan=scanno, channel=channel)
            # For the purposes of testing, limit to the first 3 mjds
            ir = ir.isel(mjd=slice(3))
            # Load the look vector and transform it to ECI
            ir['look_eci'] = get_pointing_for_ir_dataset(channel, ir, attitude['Qachieved'])
            for mjd, ir in ir.groupby('mjd'):
                for pixel in range(128):
                    pixel_look = l1.osl1.GetPixelCFUnitVectorExt(mjd, l1_inst_dict[channel], 0.0, pixel + 0.5)[1]
                    pixel_look_eci = ECI(*l1.osl1.CFToECI(mjd, pixel_look, l1.osl1.ECI_SPACECRAFT)[1])
                    np.testing.assert_allclose(ir['look_eci'][pixel], pixel_look_eci, atol=1e-9,
                                               err_msg=f'ScanNo: {scanno}', verbose=True)

    def test_pointing_for_ir_channel1_dataset(self):
        """ Test comparing the IR channel 1 look vector solution based on the attitude quaternion data to the look
            vectors returned by the c++ Level1 Services on points between the attitude solution for a dataset.
        """
        self.pointing_for_ir_channel_dataset(channel=1)

    def test_pointing_for_ir_channel2_dataset(self):
        """ Test comparing the IR channel 2 look vector solution based on the attitude quaternion data to the look
            vectors returned by the c++ Level1 Services on points between the attitude solution for a dataset.
        """
        self.pointing_for_ir_channel_dataset(channel=2)

    def test_pointing_for_ir_channel3_dataset(self):
        """ Test comparing the IR channel 3 look vector solution based on the attitude quaternion data to the look
            vectors returned by the c++ Level1 Services on points between the attitude solution for a dataset.
        """
        self.pointing_for_ir_channel_dataset(channel=3)

    def tangent_point_for_ir_dataset(self, channel: int):
        """ Test comparing the tangent point solution based on the attitude and look data with the c++ Level1 Services
            on points between the attitude solution for the IR imager.
        """
        l1_inst_dict = {1: l1.osl1.INST_IR1, 2: l1.osl1.INST_IR2, 3: l1.osl1.INST_IR3}
        for scanno in self.scan_list:
            attitude = open_level1_attitude(Orbit(scanno // 1000), scan=scanno, load_pointing=True)
            ir = open_level1_ir(Orbit(scanno // 1000), scan=scanno, channel=channel)
            # For the purposes of testing, limit to the first 3 mjds
            ir = ir.isel(mjd=slice(3))
            # Load the look vector and transform it to ECI
            ir['look_eci'] = get_pointing_for_ir_dataset(channel, ir, attitude['Qachieved'])
            ir['position_eci'] = get_position_for_ir_dataset(ir, attitude)
            ir.merge(get_tangent_point_data_for_ir_dataset(channel, ir, attitude), inplace=True)
            for mjd, ir in ir.groupby('mjd'):
                position_eci = ECI(*l1.osl1.GetOdinPosition(mjd)[1])
                # Assert the ECI positions are the to 5 meters
                np.testing.assert_allclose(ir['position_eci'], position_eci, atol=5)
                for pixel in range(128):
                    pixel_look = l1.osl1.GetPixelCFUnitVectorExt(mjd, l1_inst_dict[channel], 0.0, pixel + 0.5)[1]
                    pixel_look_eci = ECI(*l1.osl1.CFToECI(mjd, pixel_look, l1.osl1.ECI_SPACECRAFT)[1])
                    tangentpoint_eci = ECI(*l1.osl1.LOSTangentPoint(list(position_eci),
                                                                    list(pixel_look_eci), False)[1])
                    tangentpoint = Geodetic(*l1.osl1.ECIToGeodetic(mjd, list(tangentpoint_eci))[1])
                    if np.isfinite(float(ir['latitude'][pixel])):  # Do not compare NaNs
                        self.assertAlmostEqual(float(ir['latitude'][pixel]), tangentpoint.latitude, delta=0.01,
                                               msg=f'ScanNo: {scanno}')
                        self.assertAlmostEqual(float(ir['longitude'][pixel]), tangentpoint.longitude, delta=0.01,
                                               msg=f'ScanNo: {scanno}')
                        self.assertAlmostEqual(float(ir['altitude'][pixel]), tangentpoint.height * 1000.0, delta=5,
                                               msg=f'ScanNo: {scanno}')

    def test_tangent_point_for_ir_channel1_dataset(self):
        """ Test comparing the tangent point solution based on the attitude and look data with the c++ Level1 Services
            on points between the attitude solution for the IR channel 1.
        """
        self.tangent_point_for_ir_dataset(channel=1)

    def test_tangent_point_for_ir_channel2_dataset(self):
        """ Test comparing the tangent point solution based on the attitude and look data with the c++ Level1 Services
            on points between the attitude solution for the IR channel 2.
        """
        self.tangent_point_for_ir_dataset(channel=2)

    def test_tangent_point_for_ir_channel3_dataset(self):
        """ Test comparing the tangent point solution based on the attitude and look data with the c++ Level1 Services
            on points between the attitude solution for the IR channel 3.
        """
        self.tangent_point_for_ir_dataset(channel=3)

    def test_scans_in_orbit(self):
        """ Test the service returning scans in an orbit.
        """
        self.assertListEqual([scan for scan in get_scans_in_orbit_below_sza(Orbit(6432), sza=90.0)],
                             [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24,
                              25, 26, 27, 28, 63, 64, 65])

    def test_no_scans_in_orbit(self):
        """ Test the service returning scans in an orbit raises StopIteration when there are no scans.
        """
        with self.assertRaises(StopIteration):
            next(get_scans_in_orbit_below_sza(Orbit(6000), sza=90.0))
